/*
 * Copyright (c) 2003-2006 Alexandre Ratchov <alex@caoua.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 	- Redistributions of source code must retain the above
 * 	  copyright notice, this list of conditions and the
 * 	  following disclaimer.
 *
 * 	- Redistributions in binary form must reproduce the above
 * 	  copyright notice, this list of conditions and the
 * 	  following disclaimer in the documentation and/or other
 * 	  materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "hex.h"

void
hex_error(struct hex *hex, char *msg) {
	fprintf(stderr, "%u.%u: %s\n", hex->line, hex->col, msg);
}

unsigned
hex_getc(struct hex *hex, int *res) {
	int c;

	c = fgetc(hex->file);
	if (c == EOF && !feof(hex->file)) {
		hex_error(hex, strerror(errno));
		return 0;
	}
	if (c == '\n') {
		hex->line++;
		hex->col = 1;
	} else {
		hex->col++;
	}
	*res = c;
	return 1;
}

unsigned
hex_nibble(struct hex *hex, unsigned *res) {
	int c;

	if (!hex_getc(hex, &c)) {
		return 0;
	}
	if (c >= '0' && c <= '9') {
		*res = c - '0';
		return 1;
	}
	if (c >= 'A' && c <= 'F') {
		*res = 0xa + c - 'A';
		return 1;
	}
	if (c >= 'a' && c <= 'f') {
		*res = 0xa + c - 'a';
		return 1;
	}
	hex_error(hex, "bad hex digit");
	return 0;
}

unsigned
hex_byte(struct hex *hex, unsigned *res) {
	unsigned n0, n1;

	if (!hex_nibble(hex, &n0) || !hex_nibble(hex, &n1)) {
		return 0;
	}
	*res = (n0 << 4) + n1;
	return 1;
}

unsigned
hex_word(struct hex *hex, unsigned *res) {
	unsigned b0, b1;

	if (!hex_byte(hex, &b0) || !hex_byte(hex, &b1)) {
		return 0;
	}
	*res = (b0 << 8) + b1;
	return 1;
}

unsigned
hex_record(struct hex *hex, unsigned *len, unsigned *addr, 
	   unsigned *type, unsigned char *data) {
	int c;
	unsigned i, sum, val;

	if (!hex_getc(hex, &c)) {
		return 0;
	}
	if (c != ':') {
		hex_error(hex, "':' expected");
		return 0;
	}
	if (!hex_byte(hex, len)) {
		return 0;
	}
	sum = *len;
	if (!hex_word(hex, addr)) {
		return 0;
	}
	sum += (*addr & 0xff) + (*addr >> 8);
	if (!hex_byte(hex, type)) {
		return 0;
	}
	sum += *type;
	if (*type == HEX_DATA) {
		for (i = 0; i < *len; i++) {
			if (!hex_byte(hex, &val)) {
				return 0;
			}
			data[i] = val;
			sum += val;
		}
	} else if (*type != HEX_EOF) {
		hex_error(hex, "unsupported record type\n");
		return 0;
	}
	if (!hex_byte(hex, &val)) {
		return 0;
	}
	sum = (1 + ~sum) & 0xff;
	if (val != sum) {
		hex_error(hex, "checksum mismatch");
		return 0;
	}
	if (!hex_getc(hex, &c)) {
		return 0;
	}
	if (c == '\r') {
		if (!hex_getc(hex, &c)) {
			return 0;
		}
	}
	if (c != '\n') {
		hex_error(hex, "newline expected");
		return 0;
	}
	return 1;
}

void
hex_init(struct hex *hex, FILE *f) {
	hex->file = f;
	hex->line = hex->col = 1;
}

void
hex_done(struct hex *hex) {
}

