/*
 * Copyright (c) 2003-2006 Alexandre Ratchov <alex@caoua.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 	- Redistributions of source code must retain the above
 * 	  copyright notice, this list of conditions and the
 * 	  following disclaimer.
 *
 * 	- Redistributions in binary form must reproduce the above
 * 	  copyright notice, this list of conditions and the
 * 	  following disclaimer in the documentation and/or other
 * 	  materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "pic.h"
#include "hex.h"

#define SAMPLE_SIZE 64

/*
 * dump a memory block on stderr
 */
void
dump(unsigned char *data, unsigned size)
{
	unsigned i;

	i = 0;
	for (i = 0; i < size; i++) {
		if (i % 16 == 0) {
			fprintf(stderr, "%06x: ", i);
		}
		fprintf(stderr, "%02x", data[i] & 0xff);
		fprintf(stderr, (i % 16 == 15 || i + 1 == size) ? "\n" : " ");
	}
}


void
test(struct pic *pic)
{
	pic_reset(pic);
	printf("mclk = 0\n");
	getchar();

	pic_program(pic);
	printf("mclk = 1\n");
	getchar();

	pic_set(pic, 0);
	printf("rb6 = 0, rb7 = 0\n");
	getchar();

	pic_set(pic, 1);
	printf("rb6 = 0, rb7 = 1\n");
	getchar();

	pic_set(pic, PIC_CLOCK);
	printf("rb6 = 1, rb7 = 0\n");
	getchar();

	pic_set(pic, 1 | PIC_CLOCK);
	printf("rb6 = 1, rb7 = 1\n");
	getchar();
}

void
testprog(struct pic *pic)
{
	unsigned i;
	unsigned char sample[SAMPLE_SIZE], data[SAMPLE_SIZE];

	for (i = 0; i < SAMPLE_SIZE; i++) {
		sample[i] = rand() % 0x3f;
	}

	pic_reset(pic);
	pic_program(pic);
	pic_progerase(pic);

	pic_reset(pic);
	pic_program(pic);
	pic_progwrite(pic, sample, SAMPLE_SIZE);

	pic_reset(pic);
	pic_program(pic);
	pic_progread(pic, data, SAMPLE_SIZE);
	dump(data, SAMPLE_SIZE);

	for (i = 0; i < SAMPLE_SIZE; i++) {
		if (sample[i] != data[i]) {
			fprintf(stderr, "%04x: sample(%02x) != data(%02x)\n", 
				i, sample[i], data[i]);
		}
	}
}

void
testdata(struct pic *pic)
{
	unsigned i;
	unsigned char sample[SAMPLE_SIZE], data[SAMPLE_SIZE];
	
	for (i = 0; i < SAMPLE_SIZE; i++) {
		sample[i] = rand() % 0xff;
	}

	pic_reset(pic);
	pic_program(pic);
	pic_dataerase(pic);

	pic_reset(pic);
	pic_program(pic);
	pic_datawrite(pic, sample, SAMPLE_SIZE);

	pic_reset(pic);
	pic_program(pic);
	pic_dataread(pic, data, SAMPLE_SIZE);
	dump(data, SAMPLE_SIZE);

	for (i = 0; i < SAMPLE_SIZE; i++) {
		if (sample[i] != data[i]) {
			fprintf(stderr, "%04x: sample(%02x) != data(%02x)\n", 
				i, sample[i], data[i]);
		}
	}
}

/*
 * the following addresses in the HEX file are
 * mapped to the origin of prog, data, conf memories of
 * the device.
 */
unsigned prog_addr = 0;
unsigned data_addr = 0x2000;
unsigned conf_addr = 0x4000;

#define DATA_MAXSIZE 0x64
#define PROG_MAXSIZE (0x2000 * 2)
#define CONF_MAXSIZE (0x8 * 2)

/*
 * buffers where HEX file contents are stored
 */
unsigned char prog_buf[PROG_MAXSIZE];
unsigned char data_buf[DATA_MAXSIZE];
unsigned char conf_buf[CONF_MAXSIZE];

unsigned prog_size = 0, data_size = 0, conf_size = 0;

void
loadhex(char *fname)
{
#ifdef DEBUG
	unsigned i;
#endif
	FILE *f;
	struct hex hex;
	unsigned len, addr, type;
	unsigned char buf[0x100];

	if (strcmp(fname, "-") == 0) {
		f = stdin;
	} else {
		f = fopen(fname, "r");
		if (f == NULL) {
			perror(fname);
			exit(1);
		}
	}

	hex_init(&hex, f);
	prog_size = data_size = conf_size = 0;
	memset(prog_buf, '\0', PROG_MAXSIZE);
	memset(data_buf, '\0', DATA_MAXSIZE);
	memset(conf_buf, '\0', CONF_MAXSIZE);

	for (;;) {
		if (!hex_record(&hex, &len, &addr, &type, buf)) {
			exit(1);
		}
		if (type == HEX_EOF) 
			break;
#ifdef DEBUG
		fprintf(stderr, "%04x %02x: ", addr, len);
		for (i = 0; i < len; i++) {
			fprintf(stderr, "%02x ", buf[i]);
		}
		fprintf(stderr, "\n");
#endif

		if (addr >= prog_size + prog_addr && 
		    addr + len <= prog_addr + PROG_MAXSIZE) {
			memcpy(prog_buf + addr - prog_addr, buf, len);
			prog_size = addr - prog_addr + len;
		} else if (addr >= data_size + data_addr && 
			   addr + len <= data_addr + DATA_MAXSIZE) {
			memcpy(data_buf + addr - data_addr, buf, len);
			data_size = addr - data_addr + len;
		} else if (addr >= conf_size + conf_addr && 
			   addr + len <= conf_addr + CONF_MAXSIZE) {
			memcpy(conf_buf + addr - conf_addr, buf, len);
			conf_size = addr - conf_addr + len;
		} else {
			fprintf(stderr, "record at %04x-%04x out of order\n", 
				addr, addr + len);
			exit(1);
		}
	}
	hex_done(&hex);
	fclose(f);
}


int 
main(int argc, char **argv)
{
	int ch;
	struct pic pic;
	char *device, *hexfile;
	unsigned erase, verbose;

	erase = verbose = 0;
	device = hexfile = NULL;
	while ((ch = getopt(argc, argv, "vef:w:h")) != -1) {
		switch (ch) {
		case 'v':
			verbose = 1;
		case 'e':
			erase = 1;
			break;
		case 'f':
			device = optarg;
			break;
		case 'w':
			hexfile = optarg;
			break;
		default:
			goto err;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc >= 1 || device == NULL || (!erase && !hexfile)) {
	err:
		fputs("usage: picprog -f ttydev [-e] [-w hexfile]\n", stderr);
		return 0;
	}

	if (hexfile) {
		loadhex(hexfile);
		fprintf(stderr, "program: %u bytes,  data %u bytes, conf: %u bytes\n", 
			prog_size, data_size, conf_size);
		if (verbose) {
			dump(prog_buf, prog_size);
			dump(data_buf, data_size);
			dump(conf_buf, conf_size);
		}
	}
	if (!pic_init(&pic, device)) {
		exit(1);
	}
	if (erase || hexfile) {
		pic_reset(&pic);
		pic_program(&pic);
		pic_unprotect(&pic);

		pic_reset(&pic);
		pic_program(&pic);
		pic_progerase(&pic);

		if (hexfile) {
			pic_reset(&pic);
			pic_program(&pic);
			pic_progwrite(&pic, prog_buf, prog_size);
		}

		pic_reset(&pic);
		pic_program(&pic);
		pic_dataerase(&pic);

		if (hexfile) {
			pic_reset(&pic);
			pic_program(&pic);
			pic_datawrite(&pic, data_buf, data_size);

			pic_reset(&pic);
			pic_program(&pic);
			pic_cfwrite(&pic, conf_buf, conf_size);
		}
	}
	pic_done(&pic);

	return 0;
}
