/*
 * Copyright (c) 2003-2006 Alexandre Ratchov <alex@caoua.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 	- Redistributions of source code must retain the above
 * 	  copyright notice, this list of conditions and the
 * 	  following disclaimer.
 *
 * 	- Redistributions in binary form must reproduce the above
 * 	  copyright notice, this list of conditions and the
 * 	  following disclaimer in the documentation and/or other
 * 	  materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include "pic.h"

/*
 * combine two bytes into a 16bit word
 */
unsigned
data_to_word(unsigned char *data) {
	return data[0] + (data[1] << 8);
}

/*
 * split a 16bit word, in two bytes
 */
void
word_to_data(unsigned char *data, unsigned word) {
	data[0] = word & 0xff;
	data[1] = word >> 8;
}

/* 
 * writes a stop bit, then the 14bit word; and then the stop bit
 */
void
pic_writeword(struct pic *pic, unsigned val) {
	unsigned i, data;
	
	pic_set(pic, PIC_CLOCK);		/* data = 0, clk = 1 */
	pic_set(pic, 0);			/* data = 0, clk = 0 */
 
	for (i = 0; i < 14; i++) {
		data = (val & 1) ? PIC_DATA : 0;
		pic_set(pic, data | PIC_CLOCK);		/* data, clk = 1 */
		pic_set(pic, data);			/* data, clk = 0 */
		val >>= 1;
	}
	
	pic_set(pic, PIC_CLOCK);		/* data = 0, clk = 1 */
	pic_set(pic, 0);			/* data = 0, clk = 0 */
}

/*
 * read a 14bit word and ensures that start and stop bits are set to 1
 * (at high impedance)
 */
unsigned
pic_readword(struct pic *pic) {
	unsigned i, val, start, stop;
	
	val = 0;
	pic_set(pic, PIC_DATA | PIC_CLOCK);	/* data = 1, clk = 1 */
	pic_set(pic, PIC_DATA);			/* data = 1, clk = 0 */
	start = pic_get(pic);			/* read start bit */
	
	for (i = 0; i < 14; i++) {
		pic_set(pic, PIC_DATA | PIC_CLOCK);	/* data = 1, clk = 1 */
		pic_set(pic, PIC_DATA);			/* data = 1, clk = 0 */
		val |= pic_get(pic) << i;		/* read data bit */
	}

	pic_set(pic, PIC_DATA | PIC_CLOCK);	/* data = 1, clk = 1 */
	pic_set(pic, PIC_DATA);			/* data = 1, clk = 0 */
	stop = pic_get(pic);			/* read start bit */
	
	/* start bit and stop bit should be 1 */
	if (!start || !stop) {
		fputs("pic_readword: read error: start/stop bits != 0\n", stderr);
	}	
	return val;
}

/*
 * write a 6bit pic command
 */
void 
pic_command(struct pic *pic, unsigned cmd) {
	unsigned i, data;

	for (i = 0; i < 6; i++) {
		data = (cmd >> i) & 1;
	   	pic_set(pic, data | PIC_CLOCK);		/* write bit */
		pic_set(pic, data);			/* clock = 0 */
	}	
	pic_delay(pic);
}

/*
 * read data memory
 */
void
pic_dataread(struct pic *pic, unsigned char *data, unsigned size) {
	unsigned i;

	fputs("reading data memory...\n", stderr);
	for (i = 0; i < size; i++) {
		pic_command(pic, PIC_READDATA);
		data[i] = pic_readword(pic) & 0xff;
#ifdef DEBUG
		fprintf(stderr, "pic_dataread: data[%04x] = 0x%02x\n", i, data[i]);
#endif
		pic_command(pic, PIC_INCRADDR);
	}
}

/*
 * write data memory
 */
void
pic_datawrite(struct pic *pic, unsigned char *data, unsigned size) {
	unsigned i, val, verif;
	
	fputs("writing/verifying data memory...\n", stderr);
	for (i = 0; i < size; i++) {
		val = data[i];
		pic_command(pic, PIC_LOADDATA);
		pic_writeword(pic, val);
		pic_command(pic, PIC_BEGINWRITE);

		pic_command(pic, PIC_READDATA);
 		verif = pic_readword(pic) & 0xff;
		if (verif != val) {
			fprintf(stderr, 
				"pic_datawrite: at %04x: wrote %02x, verified %02x\n",
				i, val, verif);
		}
		pic_command(pic, PIC_INCRADDR);
	}
}

/*
 * erase data memory
 */
void
pic_dataerase(struct pic *pic) {
	fputs("erasing data memory...\n", stderr);
	pic_command(pic, PIC_LOADDATA);
	pic_writeword(pic, 0x3fff);
	pic_command(pic, PIC_ERASEDATA);
	pic_command(pic, PIC_BEGINERASEWRITE);
}

/*
 * read program memory
 */
void
pic_progread(struct pic *pic, unsigned char *data, unsigned size) {
	unsigned i, val;

	i = 0;
	fputs("reading program memory...\n", stderr);
	for (i = 0; i + 1 < size; i += 2) {
		pic_command(pic, PIC_READPROG);
		val = pic_readword(pic);
		pic_command(pic, PIC_INCRADDR);
		word_to_data(&data[i], val);
#ifdef DEBUG
		fprintf(stderr, 
			"pic_progread: data[%04x] = %02x %02x\n", 
			i, data[i], data[i + 1]);
#endif
	}
	if (i != size) {
		fprintf(stderr, "WARNING: prog memory size isn't even\n");
	}
}

/*
 * write program memory
 */
void
pic_progwrite(struct pic *pic, unsigned char *data, unsigned size) {
	unsigned i, val, verif;
	
	fputs("writing/verifying program memory...\n", stderr);
	for (i = 0; i + 1 < size; i += 2) {
		val = data_to_word(&data[i]);
		pic_command(pic, PIC_LOADPROG);
		pic_writeword(pic, val);
		pic_command(pic, PIC_BEGINWRITE);
#ifdef DEBUG
		fprintf(stderr, 
			"pic_progwrite: data[%04x] = %02x %02x\n", 
			i, data[i], data[i + 1]);
#endif
		pic_command(pic, PIC_READPROG);
 		verif = pic_readword(pic);
		if (verif != val) {
			fprintf(stderr, 
				"pic_progwrite: at %04x: wrote %04x, verified %04x\n",
				i, val, verif);
		}
		pic_command(pic, PIC_INCRADDR);
	}
}

/*
 * erase program memory
 */
void
pic_progerase(struct pic *pic) {
	fputs("erasing program memory...\n", stderr);
	pic_command(pic, PIC_LOADPROG);
	pic_writeword(pic, 0x3fff);
	pic_command(pic, PIC_ERASEPROG);
	pic_command(pic, PIC_BEGINWRITE);
}

/* 
 * write config memory (4 IDs, conf word)
 */
void
pic_cfwrite(struct pic *pic, unsigned char *data, unsigned size) {
	unsigned i, val, verif;
	
	fputs("writing/verifying configuration...\n", stderr);
	pic_command(pic, PIC_LOADCONF);
	pic_writeword(pic, 0);		/* unused, but required */
	for (i = 0; i + 1 < size; i += 2) {
		val = data_to_word(&data[i]);
		pic_command(pic, PIC_LOADPROG);
		pic_writeword(pic, val);
		pic_command(pic, PIC_BEGINERASEWRITE);

		pic_command(pic, PIC_READPROG);
 		verif = pic_readword(pic);
		if (verif != val) {
			fprintf(stderr, 
				"pic_cfwrite: at %04x: wrote %04x, verified %04x\n",
				i, val, verif);
		}
		pic_command(pic, PIC_INCRADDR);
	}
}

/*
 * read config memory
 */
void
pic_cfread(struct pic *pic, unsigned char *data, unsigned size) {
	unsigned i, val;

	fputs("reading config memory...\n", stderr);
	pic_command(pic, PIC_LOADCONF);
	pic_writeword(pic, 0);		/* unused but necessary */
	for (i = 0; i + 1 < size; i += 2) {
		pic_command(pic, PIC_READPROG);
		val = pic_readword(pic);
		pic_command(pic, PIC_INCRADDR);
		word_to_data(&data[i], val);
#ifdef DEBUG
		fprintf(stderr, 
			"pic_cfread: data[%04x] = %02x %02x\n", 
			i, data[i], data[i + 1]);
#endif
	}
	if (i != size) {
		fprintf(stderr, "WARNING: conf memory size isn't even\n");
	}
}

/* 
 * disable code protection (note: erases memory)
 */
void
pic_unprotect(struct pic *pic) {
	unsigned i;

	fputs("disabling code protection...\n", stderr);
	pic_command(pic, PIC_LOADCONF);
	pic_writeword(pic, 0x3fff);
	for (i = 0; i < 7; i++)
		pic_command(pic, PIC_INCRADDR);
	pic_command(pic, 1);
	pic_command(pic, 7);
	pic_command(pic, PIC_BEGINWRITE);
	pic_command(pic, 1);
	pic_command(pic, 7);
}


